// "Plaid.java" version 1.0
//
// Copyright 1997-2004 Thomas Insel <tinsel@tinsel.org>
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. The name of the author may not be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// DISCLAIMER
//
// This software is provided by the copyright holders and contributors
// "as is" and any express or implied warranties, including, but not
// limited to the implied warranties of merchantability and fitness
// for a particular purpose are disclaimed.
//
// In no event shall the copyright holders or contributors  be liable
// for any direct, indirect, special, exemplary, or consequential
// damages (including, but not limited to, procurement of substitute
// goods or services; loss of use, data, or profits; or business
// interruption) however caused an on any theory of liability, whether
// in contract, strict liability, or tort (including negligence or
// otherwise) arising in any way out of the use of this software, even
// if advised of the possibility of such damage.

// Please contact the author if you wish to distribute this file under
// different conditions or have changes to the file.
//
// You may obtain the current version of this file from
// http://www.tinsel.org/tinsel/Java/Tartan/

import java.awt.*;
import java.awt.image.*;
import java.util.StringTokenizer;

public class Plaid extends Canvas {

        private Image tartan;
        private int threadCount=0;

        private String sett;
        private int symmetry;
        private boolean darkened;
        
        // NOSYMM -- the threads are used as given in the sett.
        public static final int NOSYMM = 0;

        // SINGLEPIVOT -- the middle bands are reflected around the rightmost band.  For
        // example, "R 1 O 1 Y 1 G 1" becomes "R 1 O 1 Y 1 G 1 Y 1 O 1".
        public static final int SINGLEPIVOT = 1;

        // DOUBLEPIVOT -- the first TWO bands are not repeated.  For example,
        // "R 1 O 1 Y 1 G 1" becomes "R 1 O 1 Y 1 G 1 Y 1".  I'm not positive this is
        // the proper meaning of double pivot.
        public static final int DOUBLEPIVOT = 2;

        // Create the component with given parameters. 
        // sett is 
        
        // @param sett an alternating list of color abbreviations and thread counts
	// @param symmetry one of: NOSYMM, SINGLEPIVOT, or DOUBLEPIVOT
	// @param darkened do we darken the colors?
        
        public Plaid(String sett, int symmetry, boolean darkened) {

                sett = spaceify(sett);

                StringTokenizer st = new StringTokenizer(sett);
                int bandColors[] = new int[st.countTokens()];
                int bandThreads[] = new int[st.countTokens()]; // we leave room for symmetry
                int bandCount=0;

                // Store these for future reference.
                this.sett = sett;
                this.symmetry = symmetry;
                this.darkened = darkened;

                // Convert string to bands (a list of colors and color counts).
                while (st.countTokens() > 1) {
                        int color = string2RGB(st.nextToken());
                        int count = java.lang.Integer.valueOf(st.nextToken()).intValue();
                        if (count > 0) {
                                bandColors[bandCount] = color;
                                bandThreads[bandCount] = count;
                                threadCount += bandThreads[bandCount++];
                        }
                }
                
                // Adjust the bands for symmetry.
                if (symmetry != NOSYMM)
                        for (int i = bandCount - ((symmetry==DOUBLEPIVOT) ? 3 : 2);
                                 i > ((symmetry==DOUBLEPIVOT) ? 0 : 1);
                                 i--) {
                                        bandColors[bandCount] = bandColors[i];
                                        bandThreads[bandCount++] = bandThreads[i];
                                        threadCount += bandThreads[i];
                                }

                // Expand bands to threads.
                int threadColors[] = new int[threadCount];
                int index = 0;
                for (int i=0; i<bandCount; i++)
                        for (int j=0; j<bandThreads[i]; j++)
                                threadColors[index++] = bandColors[i];

                // "Weave" with the threads.  This is the twill pattern we       VVHH
                // use, where H is the horizontal thread and V is the            HVVH
                // vertical thread. ------------------------------------------>  HHVV
                //                                                               VHHV
                int pix[] = new int[threadCount * threadCount];
                index = 0;
                for (int y = 0; y < threadCount; y++)
                        for (int x = 0; x < threadCount; x++)
                                pix[index++] = threadColors[ (((x%4) - (y%4) + 4) %4 > 1) ? x : y ];

                // Finally, make an image from the weaving.
                tartan = createImage(new MemoryImageSource(threadCount, threadCount, pix, 0, threadCount));
        }

        public void paint(Graphics g) {
                g.drawImage(tartan, 0, 0, null);
        }

        public Dimension preferredSize() {
                return new Dimension(threadCount, threadCount);
        }

        public String toString() {
                return getClass().getName() +
                        "[sett=" + sett + ",symmetry=" + symmetry + (darkened ? ",darkened" : "") + "]";
        }

        //
        // --- Color Utility Functions
        //	
        //     Since we work with integers, we do these functions on our own instead of using Color.*.
        //

        private int rgb(int r, int g, int b) {

                if (darkened) { // this algorithm is from XTartan.
                        if ((r==b && b==g) || (r==g && b==0) ||(b==0 && g==0)) { // gray, yellow, red
                                r = (r*2)/3;
                                g = (g*2)/3;
                                b = (b*2)/3;
                        } else { // other colors
                                r /= 2;
                                b = (b*2)/5;
                                g /= 2;
                        }
                }

                return (255<<24) + (r<<16) + (g<<8) + b;
        }

        private int string2RGB(String s) {
                // I have copied these color abbreviations from XTartan, and the RGB values
                // from /usr/X11/lib/X11/rgb.txt.  I think that they're too bright as is, but
                // look pretty good when darkened.

                if      (s.equalsIgnoreCase("r"))  return rgb(255,0,0);    // Red
                else if (s.equalsIgnoreCase("cr")) return rgb(178,34,34);  // Crimson (Firebrick)
                else if (s.equalsIgnoreCase("mn")) return rgb(176,48,96);  // Maroon
                else if (s.equalsIgnoreCase("or")) return rgb(255,165,0);  // Orange
                else if (s.equalsIgnoreCase("y"))  return rgb(255,255,0);  // Yellow
                else if (s.equalsIgnoreCase("g"))  return rgb(34,139,34);  // Green (Forest Green)
                else if (s.equalsIgnoreCase("lg")) return rgb(152,251,152);// Light Green (Pale Green)
                else if (s.equalsIgnoreCase("db")) return rgb(0,0,128);    // Dark Blue (Navy)
                else if (s.equalsIgnoreCase("b"))  return rgb(0,0,205);    // Medium Blue
                else if (s.equalsIgnoreCase("az")) return rgb(135,206,235);// Azure (Sky Blue)
                else if (s.equalsIgnoreCase("pu")) return rgb(221,160,221);// Purple (Plum)
                else if (s.equalsIgnoreCase("lil"))return rgb(218,112,214);// Lilac (Orchid)
                else if (s.equalsIgnoreCase("lv")) return rgb(230,230,250);// Lavendar
                else if (s.equalsIgnoreCase("ma")) return rgb(255,0,255);  // Magenta
                else if (s.equalsIgnoreCase("br")) return rgb(165,42,42);  // Brown
                else if (s.equalsIgnoreCase("w"))  return rgb(255,255,255);// White
                else if (s.equalsIgnoreCase("gy")) return rgb(190,190,190);// Gray
                else return rgb(0,0,0); // K, BK for black
                
                // Cy: Cyan  Cor: Coral  ForGr: ForestGreen  SlB: Slate Blue
                // Glr: Goldenrod Mar: Maroon  Trq: Turquoise  Wh: Wheat
        }

        //
        // --- String Utility Functions
        ///

        // Get rid of garbage and extra white space, and spaces the letter and number sections.

        private String spaceify(String s) {
                int LETTER = 1;
                int NUMBER = 2;
                int OTHER  = 3;

                int kind, lastKind=OTHER;

                StringBuffer t = new StringBuffer(2 * s.length());

                for (int i = 0; i < s.length(); i++) {
                        char c = s.charAt(i);
                        if ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz".indexOf(c) != -1)
                                kind = LETTER;
                        else if ("0123456789".indexOf(c) != -1)
                                kind = NUMBER;
                        else
                                kind = OTHER;

                        if ( kind != OTHER && (kind == lastKind || lastKind == OTHER))
                                t.append(c);
                        else if (kind == OTHER) {
                                if (lastKind != OTHER) t.append(' ');
                        } else
                                t.append(' ').append(c);

                        lastKind = kind;
                }

                return t.toString();
        }
        
}